/*
 * Code to fetch data over a TCP link using HTTP
 * 2002-2003 HShivanan
 * 
 * For licensing terms, see the file COPYING
 * For contact information, see the file AUTHORS
 */
#include <stdio.h>
#include <string.h>
#include "net.h"
#include "fn.h"
#include "stuff.h"
#include "connection.h"
#include "stream.h"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

/* The following two functions were "borrowed" (ok, stolen)
 * from the WGet source code and modified slightly.
 */

static int select_fd (int fd, int maxtime)
{
  fd_set fds, exceptfds;
  struct timeval timeout;

  FD_ZERO (&fds);
  FD_SET (fd, &fds);
  FD_ZERO (&exceptfds);
  FD_SET (fd, &exceptfds);
  timeout.tv_sec = maxtime;
  timeout.tv_usec = 0;
  /* HPUX reportedly warns here.  What is the correct incantation?  */
	return select (fd+1,&fds,NULL,
		 &exceptfds, &timeout);
}


/* Keep applying select until we timeout or have something to read */
static int n_read (int fd, char *buf, int len,int timeout)
{
  int res;

  do
    {
	  do
	  {
	      res = select_fd (fd, timeout);
	  } while (res == -1 && errno == EINTR);
	  if (res <= 0)
	  {
	      if (res == 0)
	      errno = ETIMEDOUT;
	      return res;
	   }
      res = n_recieve(fd, buf, len);
    } while (res == -1 && errno == EINTR);

  return res;
}

/* Make sure the requested file was fetched properly*/
/* TODO: Do a more robust check. 30[12] Location Changed fails*/
static int validate_http_header(char *buffer)
{
	verbose("Validating HTTP Header...\n");
	if (strncmp(buffer,"HTTP/",5))
		return 0;
	if (strncmp(buffer+9,"200",3))
		return 0;
	verbose("HTTP Buffer validated...OK\n");
	return 1;
}


/*Constructor and destructor for an HTTP REQUEST structure*/
HTTP_REQUEST * new_http_request()
{
	HTTP_REQUEST *hr = (HTTP_REQUEST *)fn_malloc(sizeof(HTTP_REQUEST));
	hr->agent = NULL;
	hr->connection = NULL;
	hr->version = NULL;
	return hr;	
}
void free_http_request(HTTP_REQUEST *hr)
{
//	if (hr->agent)
//		fn_free(hr->agent);
//	if (hr->connection)
//		fn_free(hr->connection);
//	if (hr->host)
//		fn_free(hr->host);
//	if (hr->version)
//		fn_free(hr->version);
	fn_free(hr);
}

/* some web servers are picky and require \r\n at the end, not just
 * \n.
 */
static STREAM *construct_header(HTTP_REQUEST *hr,char *file)
{
	int l;
	STREAM *stream;
	/* TODO: Calculate the buffer size and allocate memory for it.*/
	char buffer[1024];
	stream = stream_create(1024);
	l = sprintf(buffer,"GET %s HTTP/%s\r\n",
			file,
			hr->version?hr->version:"1.0" //default to version 1.0
			);
	stream_write(stream,buffer,l);
	
	/*stream_write(stream,"GET ",4);
	stream_write(stream,file,strlen(file));
	stream_write(stream," HTTP/",6);
	ver = hr->version?hr->version:"1.0";
	stream_write(stream,ver,strlen(ver));
	stream_write(stream,"\r\n\0",3);*/
	
	if (hr->host)
	{
		l = sprintf(buffer,"Host: %s\r\n",hr->host);
		stream_write(stream,buffer,l);
	}
	if (hr->agent)
	{
		l = sprintf(buffer,"User-Agent: %s\r\n",hr->agent);
		stream_write(stream,buffer,l);
	}
	if (hr->connection)
	{
		l = sprintf(buffer,"Connection: %s\r\n",hr->connection);
		stream_write(stream,buffer,l);
	}

	
	/*Sometimes, we get problems if the request is null-terminated.
	 * However, because we need to print out the request in
	 * verbose mode, we append the null-character, but remember
	 * not to send it to the server by sending only size-1 bytes
	 */
	stream_write(stream,"\r\n\0",3);
	return stream;	
}

/* Construct an HTTP request header for use with a proxy server */
static STREAM * construct_request_proxy(FN_OPTIONS *fn,char *channel)
{
	fn->http_request->host = NULL;
	return construct_header(fn->http_request,channel);
}

/*Construct an HTTP request header and return it as a text stream*/
static STREAM * construct_request(FN_OPTIONS *fn,char *channel)
{
	char *url;
	char *file;
	STREAM *stream;
	int i,len,url_index=0;
	len = strlen(channel);
	i = 0;

	/*get i to point to the file part of an url */
	while (1)
	{
		char ch;
		if (i>len)
			break;
		ch = channel[i];
		if (ch=='/')
		{
			if (channel[i+1]=='/')
			{
				i+=2;
				continue;
			}
			else
				break;
		}
		i++;
	}
	if (i >= len || channel[i]!='/')
	{
		error ("Invalid Url:%s\n",channel);
		return NULL;
	}
	url = (char *)fn_malloc(i+1);
	memcpy(url,channel,i);
	url[i] = 0;
	file = (char *)fn_malloc(len-i+1);
	memcpy(file,channel+i,len-i);
	file[len-i]=0;
	if (!strncmp(url,"http://",7))
		url_index=7;

	/*We don't support SSL*/
	/*else if (!strncmp(url,"https://",8))
		url_index=8;*/
	else
		url_index=0;
	//TODO: parse out a port from the url in form
	//      address:port-number
	fn->http_request->host=url+url_index;
	stream = construct_header(fn->http_request,file);
	//fn_free(file);
	//fn_free(url);
	return stream;
}
STREAM * fetch_url(FN_OPTIONS *fn,char *url)
{
	STREAM *req_stream;
	STREAM *data;
	int l;
	int fd;
	char buffer[SOCKET_BUFFER_SIZE];
	if (fn->proxy)
	{
		req_stream = construct_request_proxy(fn,url);
		if (!req_stream)
			return NULL;
		fd = n_connect(fn->proxy,fn->proxy_port);
	}
	else
	{
		req_stream = construct_request(fn,url);
		if (!req_stream)
			return NULL;
		fd = n_connect(fn->http_request->host,80);
	}
	
	if (fd<=0)
	{
		error("Unable to contact host:%s\n",url);
		return NULL;
	}
	
	verbose("HTTP Request Header:\n");
	verbose(req_stream->buffer);
	
	
	/*IMPORTANT: send (stream_size-1) bytes because the the last
	 * NULL character should NOT be sent. This screwes up the response
	 * from some web servers (for ex, freshmeat.net)
	 */
	l = n_sendRaw (fd,req_stream->buffer,req_stream->size-1);
	verbose ("%i bytes sent...\n",l);
		
	
	//stream_free(req_stream);
	data = stream_create(SOCKET_BUFFER_SIZE);
	verbose("Reading ");
	while ((l = n_read(fd,buffer,SOCKET_BUFFER_SIZE,10))>0)
	{
		verbose (".");
		fflush (stdout);
		stream_write (data,buffer,l);
	}
	n_close (fd);
	verbose("OK\n");
	verbose("%i bytes read\n",data->size);
	
	stream_add_char (data, '\0'); 
	if (validate_http_header(data->buffer))
		return data;
	else
	{
		error("Unable to fetch document:%s\n",url);
		stream_free(data);
		return NULL;
	}
}
